import 'package:extensions_pack/extensions_pack.dart';
import 'package:flutter/material.dart';
import 'package:flutter_spinkit/flutter_spinkit.dart';
import 'package:widgets_pack/buttons/buttons.dart';

enum AppButtonType {
  primary,
  secondary,
  tertiary,
  icon;

  bool get isPrimary => this == AppButtonType.primary;
  bool get isIcon => this == AppButtonType.icon;
}

class AppButton extends StatefulWidget {
  final AppButtonType type;
  final Widget child;
  final EdgeInsets? padding;
  final FutureOrCallback? onPressed;
  final FutureOrCallback? onPressedDisabled;

  const AppButton._({
    required this.type,
    required this.child,
    this.padding,
    this.onPressed,
    this.onPressedDisabled,
    super.key,
  });

  const AppButton.primary({
    required this.child,
    this.padding,
    this.onPressed,
    this.onPressedDisabled,
    super.key,
  }) : type = AppButtonType.primary;

  const AppButton.secondary({
    required this.child,
    this.padding,
    this.onPressed,
    this.onPressedDisabled,
    super.key,
  }) : type = AppButtonType.secondary;

  const AppButton.tertiary({
    required this.child,
    this.padding,
    this.onPressed,
    this.onPressedDisabled,
    super.key,
  }) : type = AppButtonType.tertiary;

  factory AppButton.icon({
    required Widget icon,
    EdgeInsets? padding,
    FutureOrCallback? onPressed,
    FutureOrCallback? onPressedDisabled,
    Color? backgroundColor,
    String? tooltip,
    Key? key,
  }) {
    return _AppIconButton(
      key: key,
      backgroundColor: backgroundColor,
      padding: padding,
      onPressed: onPressed,
      onPressedDisabled: onPressedDisabled,
      tooltip: tooltip,
      child: icon,
    );
  }

  factory AppButton.fromConfig(ButtonConfig config) {
    return switch (config.type) {
      (ButtonType.elevated) => AppButton.primary(
          onPressed: config.onPressed,
          child: config.child,
        ),
      (ButtonType.outlined) => AppButton.secondary(
          onPressed: config.onPressed,
          child: config.child,
        ),
      (ButtonType.text) => AppButton.tertiary(
          onPressed: config.onPressed,
          child: config.child,
        ),
      (ButtonType.icon) => _AppIconButton(
          onPressed: config.onPressed,
          child: config.child,
        ),
    };
  }

  @override
  State<AppButton> createState() => _AppButtonState();
}

class _AppButtonState extends State<AppButton> {
  final _loadingNotifier = ValueNotifier<bool>(false);

  AppButtonType get _type => widget.type;

  FutureOrCallback? get _onPressed {
    if (widget.onPressed == null) {
      return null;
    }

    return () async {
      if (_loadingNotifier.value) {
        return;
      }

      _loadingNotifier.value = true;

      try {
        await widget.onPressed!();
      } catch (e) {
        rethrow;
      } finally {
        if (mounted) {
          _loadingNotifier.value = false;
        }
      }
    };
  }

  Widget get _loadingChild {
    return IntrinsicWidth(
      child: SpinKitThreeBounce(
        color: _type.isPrimary ? context.colorTheme.onPrimary : context.colorTheme.primary,
        size: 16,
      ),
    );
  }

  Widget get _child {
    return ValueListenableBuilder<bool>(
      valueListenable: _loadingNotifier,
      builder: (context, isLoading, child) {
        return Stack(
          alignment: Alignment.center,
          children: [
            Visibility(
              visible: !isLoading,
              maintainAnimation: true,
              maintainState: true,
              maintainSize: true,
              child: child!,
            ),
            if (isLoading) _loadingChild.sized(height: 16),
          ],
        );
      },
      child: widget.child,
    );
  }

  @override
  Widget build(BuildContext context) {
    return ButtonTheme.fromButtonThemeData(
      data: Theme.of(context).buttonTheme.copyWith(
            padding: widget.padding,
          ),
      child: GestureDetector(
        onTap: _onPressed == null && widget.onPressedDisabled != null ? widget.onPressedDisabled : null,
        child: switch (widget.type) {
          AppButtonType.primary => ElevatedButton(
              onPressed: _onPressed,
              child: _child,
            ),
          AppButtonType.secondary => OutlinedButton(
              onPressed: _onPressed,
              child: _child,
            ),
          AppButtonType.tertiary => TextButton(
              onPressed: _onPressed,
              child: _child,
            ),
          AppButtonType.icon => throw UnimplementedError(),
        },
      ),
    );
  }
}

class _AppIconButton extends AppButton {
  final String? tooltip;
  final Color? backgroundColor;

  const _AppIconButton({
    required Widget child,
    EdgeInsets? padding,
    FutureOrCallback? onPressed,
    FutureOrCallback? onPressedDisabled,
    this.backgroundColor,
    this.tooltip,
    super.key,
  }) : super._(
          type: AppButtonType.icon,
          child: child,
          padding: padding,
          onPressed: onPressed,
          onPressedDisabled: onPressedDisabled,
        );

  @override
  State<AppButton> createState() => _AppIconsButtonState();
}

class _AppIconsButtonState extends _AppButtonState {
  _AppIconButton get _widget => widget as _AppIconButton;

  @override
  Widget get _loadingChild {
    return IntrinsicWidth(
      child: SpinKitCircle(
        color: context.colorTheme.primary,
        size: 16,
      ),
    );
  }

  @override
  Widget build(BuildContext context) {
    return GestureDetector(
      onTap: _onPressed == null && widget.onPressedDisabled != null ? widget.onPressedDisabled : null,
      child: IconButton(
        style: ButtonStyle(
          backgroundColor: MaterialStateProperty.all<Color?>(
            _widget.backgroundColor,
          ),
        ),
        onPressed: _onPressed,
        padding: widget.padding ?? EdgeInsets.zero,
        visualDensity: const VisualDensity(),
        splashRadius: 24,
        tooltip: _widget.tooltip,
        icon: _child,
      ),
    );
  }
}
