import 'package:extensions_pack/extensions_pack.dart';
import 'package:flutter/material.dart';
import 'package:widgets_pack/themes/themes.dart';

enum AppSwitchSide {
  left,
  right;

  bool get isLeft => this == AppSwitchSide.left;

  bool get isRight => this == AppSwitchSide.right;
}

class AppSwitch extends StatefulWidget {
  final Map<AppSwitchSide, String> labels;
  final Future<bool> Function(AppSwitchSide)? beforeChange;
  final AppSwitchSide? disabledSide;
  final AppSwitchSide? initialSide;
  final FutureValueChanged<AppSwitchSide>? onChanged;

  const AppSwitch({
    required this.labels,
    super.key,
    this.beforeChange,
    this.disabledSide,
    this.initialSide,
    this.onChanged,
  });

  @override
  State<AppSwitch> createState() => _AppSwitchState();
}

class _AppSwitchState extends State<AppSwitch> {
  late final _sideNotifier = ValueNotifier<AppSwitchSide>(
    widget.initialSide ?? AppSwitchSide.left,
  );

  @override
  Widget build(BuildContext context) {
    return Row(
      mainAxisSize: MainAxisSize.min,
      children: [
        if (widget.labels[AppSwitchSide.left] != null) ...[
          ValueListenableBuilder<AppSwitchSide>(
            valueListenable: _sideNotifier,
            builder: (context, side, child) {
              final isActive = side.isLeft;

              return Text(
                widget.labels[AppSwitchSide.left]!,
                style: context.textTheme.bodyLarge?.copyWith(
                  fontWeight: isActive ? FontWeight.w500 : FontWeight.w400,
                ),
              );
            },
          ),
          const Spacing(kXSSize)
        ],
        ValueListenableBuilder<AppSwitchSide>(
          valueListenable: _sideNotifier,
          builder: (context, side, child) {
            final isDisabled = side == widget.disabledSide;

            return InkWell(
              splashFactory: InkRipple.splashFactory,
              borderRadius: BorderRadius.circular(20),
              onTap: () async {
                final nextSide = side.isLeft ? AppSwitchSide.right : AppSwitchSide.left;
                final canChange = await widget.beforeChange?.call(nextSide) ?? true;

                if (!canChange) {
                  return;
                }

                await widget.onChanged?.call(nextSide);
                _sideNotifier.value = side.isLeft ? AppSwitchSide.right : AppSwitchSide.left;
              },
              child: Stack(
                children: [
                  AnimatedContainer(
                    duration: const Duration(milliseconds: 300),
                    height: 28,
                    width: 56,
                    decoration: BoxDecoration(
                      borderRadius: BorderRadius.circular(20),
                      border: Border.all(
                        color: side == widget.disabledSide ? AppColors.grey500 : context.colorTheme.primary,
                      ),
                    ),
                  ),
                  AnimatedPositioned(
                    duration: const Duration(milliseconds: 100),
                    top: 3 + (isDisabled ? 2 : 0),
                    left: side.isLeft ? 3 + (isDisabled ? 2 : 0) : (56 - 22 - 3 - (isDisabled ? 2 : 0)),
                    child: AnimatedContainer(
                      duration: const Duration(milliseconds: 100),
                      height: 22 - (isDisabled ? 4 : 0),
                      width: 22 - (isDisabled ? 4 : 0),
                      decoration: BoxDecoration(
                        shape: BoxShape.circle,
                        color: isDisabled ? AppColors.grey500 : context.colorTheme.primary,
                      ),
                    ),
                  ),
                ],
              ),
            );
          },
        ),
        if (widget.labels[AppSwitchSide.right] != null) ...[
          const Spacing(kXSSize),
          ValueListenableBuilder<AppSwitchSide>(
            valueListenable: _sideNotifier,
            builder: (context, side, child) {
              final isActive = !side.isLeft;

              return Text(
                widget.labels[AppSwitchSide.right]!,
                style: context.textTheme.bodyLarge?.copyWith(
                  fontWeight: isActive ? FontWeight.w500 : FontWeight.w400,
                ),
              );
            },
          )
        ],
      ],
    );
  }
}
