import 'dart:async';
import 'dart:ui';

import 'package:extensions_pack/extensions_pack.dart';
import 'package:flutter/material.dart';
import 'package:flutter/scheduler.dart';
import 'package:widgets_pack/inputs/inputs.dart';

typedef OnChanged<T> = FutureOr<void> Function(T? value);
typedef ChangeItemValidator<T> = FutureOr<bool> Function(T?);
typedef AddItemCallback<T> = FutureOr<T?> Function(String);

class _AppAutocompleteOption<T> {
  final String label;
  final T? value;

  const _AppAutocompleteOption({
    required this.label,
    required this.value,
  });
}

class AppAutocomplete<T> extends StatefulWidget {
  final String? labelText;
  final String? helperText;
  final ListPopulator<T> populator;
  final OnChanged<T>? onChanged;
  final T? initialItem;
  final bool isRequired;
  final bool isEnabled;
  final bool updateTextAfterOnChanged;
  final FormFieldValidator<String>? validator;
  final TextEditingController? textEditingController;
  final FocusNode? focusNode;
  final Widget Function(T)? itemTileBuilder;
  final Widget? prefixIcon;
  final bool showClearButton;
  final String? isRequiredMessage;
  final VoidCallback? onFocusGained;
  final ChangeItemValidator<T>? changeItemValidator;
  final AddItemCallback<T>? onAddItem;
  final TextInputAction? textInputAction;
  final bool triggerOnChangeOnInitialItem;
  final Duration debounceDuration;

  const AppAutocomplete({
    required this.populator,
    super.key,
    this.helperText,
    this.labelText,
    this.onChanged,
    this.initialItem,
    this.validator,
    this.textEditingController,
    this.itemTileBuilder,
    this.focusNode,
    this.prefixIcon,
    this.isRequiredMessage,
    this.onFocusGained,
    this.changeItemValidator,
    this.onAddItem,
    this.isRequired = false,
    this.isEnabled = true,
    this.updateTextAfterOnChanged = true,
    this.showClearButton = true,
    this.textInputAction,
    this.triggerOnChangeOnInitialItem = false,
    this.debounceDuration = const Duration(milliseconds: 350),
  });

  static Duration debounce = const Duration(milliseconds: 350);

  @override
  State<AppAutocomplete<T>> createState() => _AppAutocompleteState<T>();
}

class _AppAutocompleteState<T> extends State<AppAutocomplete<T>> {
  late final _debounceDuration = widget.populator is StaticListPopulator<T> ? Duration.zero : widget.debounceDuration;
  late final _textController = widget.textEditingController ?? TextEditingController();
  late final _focusNode = widget.focusNode ?? FocusNode();
  late final _debouncer = Debouncer(duration: _debounceDuration);

  final _loadingNotifier = ValueNotifier<bool>(false);
  final _inputKey = GlobalKey();

  String _asString(T item) => widget.populator.asString(item);

  @override
  void initState() {
    super.initState();

    _textController.text = widget.initialItem != null ? _asString(widget.initialItem as T) : '';

    _focusNode.addListener(() {
      if (_focusNode.hasFocus) {
        widget.onFocusGained?.call();
        _textController.text = '${_textController.text.trim()} ';
        _setState();
        _textController.text = _textController.text.trim();
        _setState();
      }
    });

    WidgetsBinding.instance.addPostFrameCallback((_) {
      if (widget.triggerOnChangeOnInitialItem && widget.initialItem != null) {
        widget.onChanged?.call(widget.initialItem);
      }
    });
  }

  @override
  void dispose() {
    super.dispose();

    if (widget.textEditingController == null) {
      _textController.dispose();
    }

    if (widget.focusNode == null) {
      _focusNode.dispose();
    }
  }

  void _setState() {
    if (mounted) setState(() {});
  }

  FutureOr<bool> _canChangeItem(T? item) async {
    return widget.changeItemValidator?.call(item) ?? true;
  }

  double _computeMaxHeightFromPosition() {
    final RenderBox inputBox = _inputKey.currentContext!.findRenderObject() as RenderBox;
    final inputPosition = inputBox.localToGlobal(Offset.zero);
    final screenHeight = MediaQuery.of(context).size.height;
    final maxHeight = screenHeight - inputPosition.dy - inputBox.size.height;

    return clampDouble(maxHeight, 0, 200);
  }

  @override
  Widget build(BuildContext context) {
    return ValueListenableBuilder<bool>(
      valueListenable: _loadingNotifier,
      builder: (context, loading, child) {
        return Stack(
          alignment: Alignment.bottomCenter,
          children: [
            child!,
            if (loading)
              Align(
                alignment: Alignment.bottomCenter,
                child: Padding(
                  padding: kSpacer.x.xs + kSpacer.bottom.value(0.1),
                  child: LinearProgressIndicator(
                    color: context.colorTheme.primary,
                  ),
                ),
              ),
          ],
        );
      },
      child: RawAutocomplete<_AppAutocompleteOption<T>>(
        textEditingController: _textController,
        focusNode: _focusNode,
        fieldViewBuilder: (context, textEditingController, focusNode, onFieldSubmitted) {
          return AppTextField(
            key: _inputKey,
            focusNode: focusNode,
            helperText: widget.helperText,
            enabled: widget.isEnabled,
            labelText: widget.labelText,
            onFieldSubmitted: (_) => onFieldSubmitted(),
            prefixIcon: widget.prefixIcon,
            suffixIcon: widget.isEnabled
                ? textEditingController.text.trim().isEmpty || !widget.showClearButton
                    ? const Icon(Icons.arrow_drop_down)
                    : GestureDetector(
                        onTap: () async {
                          if (await _canChangeItem(null)) {
                            textEditingController.clear();
                            focusNode.requestFocus();
                            widget.onChanged?.call(null);
                            _setState();
                          }
                        },
                        child: const Icon(Icons.clear),
                      )
                : null,
            controller: textEditingController,
            textInputAction: widget.textInputAction,
            validator: widget.validator,
          );
        },
        optionsViewBuilder: (context, onSelected, options) {
          if (!widget.isEnabled) return const SizedBox();

          return Transform.translate(
            offset: Offset(0, widget.helperText != null ? -16 : 0),
            child: Align(
              alignment: Alignment.topLeft,
              child: Material(
                elevation: 4.0,
                child: Container(
                  padding: const EdgeInsets.symmetric(horizontal: 4.0),
                  constraints: BoxConstraints(
                    maxHeight: _computeMaxHeightFromPosition(),
                    maxWidth: (_inputKey.currentContext!.findRenderObject() as RenderBox).size.width,
                  ),
                  decoration: BoxDecoration(
                    color: Colors.white,
                    borderRadius: const BorderRadius.vertical(
                      bottom: Radius.circular(4),
                    ),
                    boxShadow: [
                      BoxShadow(
                        blurRadius: 10.0,
                        color: Colors.black.withOpacity(0.25),
                        offset: const Offset(5.0, 10.0),
                      ),
                    ],
                  ),
                  child: ListView.separated(
                    padding: EdgeInsets.zero,
                    shrinkWrap: true,
                    itemCount: options.length,
                    separatorBuilder: (context, index) => const Divider(height: 0),
                    itemBuilder: (BuildContext context, int index) {
                      final option = options.elementAt(index);

                      return InkWell(
                        onTap: () async {
                          if (await _canChangeItem(option.value)) {
                            if (option.value == null && widget.onAddItem != null) {
                              final item = await widget.onAddItem!(_textController.text);
                              if (item != null) {
                                onSelected(
                                  _AppAutocompleteOption(
                                    label: _asString(item),
                                    value: item,
                                  ),
                                );
                              }
                            } else {
                              onSelected(option);
                            }
                          }
                        },
                        child: Builder(
                          builder: (BuildContext context) {
                            final bool highlight = AutocompleteHighlightedOption.of(context) == index;

                            if (highlight) {
                              SchedulerBinding.instance.addPostFrameCallback((Duration timeStamp) {
                                Scrollable.ensureVisible(context, alignment: 0.5);
                              });
                            }

                            return Container(
                              color: highlight ? context.colorTheme.primary.withAlpha(50) : null,
                              padding: widget.itemTileBuilder != null ? null : const EdgeInsets.all(16.0),
                              child: option.value != null
                                  ? (widget.itemTileBuilder?.call(option.value as T) ?? Text(option.label))
                                  : Text(option.label),
                            );
                          },
                        ),
                      );
                    },
                  ),
                ),
              ),
            ),
          );
        },
        displayStringForOption: (option) => option.label,
        onSelected: (option) async {
          widget.onChanged?.call(option.value);
          if (!widget.updateTextAfterOnChanged) {
            _textController.clear();
            _focusNode.unfocus();
          }
        },
        optionsBuilder: (TextEditingValue textEditingValue) async {
          if (!widget.isEnabled) return const [];

          final items = await _debouncer.run(
            () async {
              _loadingNotifier.value = true;

              return await widget.populator.getItems(
                textEditingValue.text,
              );
            },
          );

          final options = (items ?? []).map((e) {
            return _AppAutocompleteOption(
              label: _asString(e),
              value: e,
            );
          }).toList();

          if (widget.onAddItem != null && textEditingValue.text.isNotEmpty) {
            options.add(
              _AppAutocompleteOption(
                label: '+ Adicionar ${textEditingValue.text}',
                value: null,
              ),
            );
          }

          _loadingNotifier.value = false;
          _setState();

          return options;
        },
      ),
    );
  }
}

class Debouncer {
  final Duration duration;

  Debouncer({required this.duration});

  Timer? _timer;

  Future<T?> run<T>(FutureOr<T> Function() action) {
    final completer = Completer<T?>();
    if (_timer?.isActive ?? false) {
      _timer?.cancel();
    }

    _timer = Timer(
      duration,
      () async {
        completer.complete(await action());
      },
    );

    return completer.future;
  }
}
