import 'package:extensions_pack/extensions_pack.dart';
import 'package:flutter/material.dart';
import 'package:rxdart/rxdart.dart';
import 'package:widgets_pack/themes/themes.dart';

class AppTextField extends StatefulWidget {
  final AutovalidateMode? autovalidateMode;
  final TextEditingController? controller;
  final Duration debounceTime;
  final bool enabled;
  final FocusNode? focusNode;
  final String? helperText;
  final String? hintText;
  final String? initialValue;
  final String? labelText;
  final int? maxLength;
  final int? maxLines;
  final ValueChanged<String>? onChanged;
  final ValueChanged<String>? onFieldSubmitted;
  final Widget? prefixIcon;
  final Widget? suffixIcon;
  final TextInputAction? textInputAction;
  final FormFieldValidator<String>? validator;

  const AppTextField({
    this.autovalidateMode,
    this.controller,
    this.debounceTime = Duration.zero,
    this.enabled = true,
    this.focusNode,
    this.helperText,
    this.hintText,
    this.initialValue,
    this.labelText,
    this.maxLength,
    this.maxLines = 1,
    this.onChanged,
    this.onFieldSubmitted,
    this.prefixIcon,
    this.suffixIcon,
    this.textInputAction,
    this.validator,
    super.key,
  });

  @override
  State<AppTextField> createState() => _AppTextFieldState();
}

class _AppTextFieldState extends State<AppTextField> {
  late final _controller = widget.controller ?? TextEditingController();
  late final _focusNode = widget.focusNode ?? FocusNode();

  final _behaviorSubject = BehaviorSubject<String>();

  @override
  void initState() {
    super.initState();

    if (widget.initialValue != null) {
      _controller.text = widget.initialValue!;
    }

    _behaviorSubject.debounceTime(widget.debounceTime).listen(widget.onChanged);
  }

  @override
  void dispose() {
    if (widget.controller == null) {
      _controller.dispose();
    }

    if (widget.focusNode == null) {
      _focusNode.dispose();
    }

    _behaviorSubject.close();

    super.dispose();
  }

  @override
  Widget build(BuildContext context) {
    final theme = Theme.of(context);
    final textTheme = theme.textTheme;
    final textStyle = textTheme.bodyLarge!.copyWith(
      color: AppColors.black80,
      fontSize: 16,
      letterSpacing: 0.15,
    );

    return Theme(
      data: theme.copyWith(
        textTheme: textTheme.copyWith(
          bodyLarge: textStyle,
        ),
      ),
      child: TextFormField(
        autovalidateMode: widget.autovalidateMode,
        controller: _controller,
        enabled: widget.enabled,
        focusNode: _focusNode,
        decoration: InputDecoration(
          helperText: widget.helperText,
          hintText: widget.hintText,
          labelText: widget.labelText,
          prefixIcon: widget.prefixIcon?.padded(
            const EdgeInsets.symmetric(horizontal: 12),
          ),
          suffixIcon: widget.suffixIcon?.padded(
            const EdgeInsets.symmetric(horizontal: 12),
          ),
          suffixIconConstraints: BoxConstraints.loose(
            /// 24 is the size of the icon, 12's are the padding
            const Size(12 + 24 + 12, 24),
          ),
        ),
        maxLength: widget.maxLength,
        maxLines: widget.maxLines,
        onChanged: _behaviorSubject.add,
        onFieldSubmitted: widget.onFieldSubmitted,
        textInputAction: widget.textInputAction,
        validator: widget.validator,
      ),
    );
  }
}
